import head_tracker.ColoredImagePatchMatch;
import head_tracker.LocalHeadTracker;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;

import org.opencv.calib3d.Calib3d;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.Point;
import org.opencv.core.Point3;
import org.opencv.core.Size;
import org.opencv.highgui.Highgui;

import de.tu_darmstadt.fsr.headtracker.TrainingData;


public class CreateTestDatasetMain {
	
	private static LocalHeadTracker headTracker;
	private static TrainingData trainingData;
	
	private static final double[] DISTORTION_MATRIX_VALUES = new double[]{
			/*This matrix should have 5 values*/
			0.07444480392067945,
			-0.7817175834131075,
			0,
			0,
			2.65433773093283
	};
	
	private static final double[][] CAMERA_MATRIX_VALUES = new double[][]{
			/*This matrix should have 3x3 values*/
			{702.1458656346429, 0, 319.5},
			{0, 702.1458656346429, 239.5},
			{0, 0, 1}
	};
	
	public static void main(String[] args) throws ParserConfigurationException, FileNotFoundException, TransformerFactoryConfigurationError, TransformerException {
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
		createDatasetSimulator();
	}
	
	public static void createDatasetSimulator() throws FileNotFoundException, ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException{
		Mat cameraMatrix = new Mat(new Size(3,3), CvType.CV_64F);
		for(int i=0;i<CAMERA_MATRIX_VALUES.length; i++){
			cameraMatrix.put(i, 0, CAMERA_MATRIX_VALUES[i]);
		};
		
		MatOfDouble distortionMatrix = new MatOfDouble();
		distortionMatrix.fromArray(DISTORTION_MATRIX_VALUES);
		
		File xmlFile = new File("ressources/input-data.xml");
		Dataset dataset = new Dataset(new LinkedList<Picture>(), cameraMatrix, distortionMatrix);
		XmlInterface xmlInterface = new XmlInterface();
		headTracker = new LocalHeadTracker();
		trainingData = headTracker.getTrainingData();
		
		Mat basisExchangeOpenCVToEuropeanCartesian = Mat.zeros(new Size(3,3),CvType.CV_64F);
		double[] row0 = new double[]{0,0,1};
		double[] row1 = new double[]{-1,0,0};
		double[] row2 = new double[]{0,1,0};
		basisExchangeOpenCVToEuropeanCartesian.put(0, 0, row0);
		basisExchangeOpenCVToEuropeanCartesian.put(1, 0, row1);
		basisExchangeOpenCVToEuropeanCartesian.put(2, 0, row2);
		
		Mat basisExchangeEuropeanCartesianToOpenCV = basisExchangeOpenCVToEuropeanCartesian.inv();
		
		/*
		 * open-cv coordinates
		 * c = 0,-59,-22 
		 * 
		 * */
		InputPicture[] inputs = new InputPicture[]{
			/*roll test series*/
			new InputPicture("13.png", 0, 0, 0, 2, 0, 0),
			new InputPicture("15.png", 0, 0, 0, 7, 0, 0),
			new InputPicture("18.png", 0, 0, 0, 12, 0, 0),
			new InputPicture("19.png", 0, 0, 0, 17, 0, 0),
			new InputPicture("20.png", 0, 0, 0, 22, 0, 0),
			new InputPicture("21.png", 0, 0, 0, 27, 0, 0),
			new InputPicture("22.png", 0, 0, 0, 32, 0, 0),
			new InputPicture("24.png", 0, 0, 0, 37, 0, 0),
			new InputPicture("26.png", 0, 0, 0, 42, 0, 0),
			new InputPicture("27.png", 0, 0, 0, 47, 0, 0),
			
			/*pitch test series*/
			new InputPicture("32.png", 0, 0, 0, -3, -16, 0),
			new InputPicture("31.png", 0, 0, 0, -3, -11, 0),
			new InputPicture("29.png", 0, 0, 0, -3, -6, 0),
			new InputPicture("28.png", 0, 0, 0, -3, -1, 0),
			new InputPicture("33.png", 0, 0, 0, -3, 4, 0),
			new InputPicture("34.png", 0, 0, 0, -3, 9, 0),
			new InputPicture("35.png", 0, 0, 0, -3, 14, 0),
			
			/*yaw test series*/
			new InputPicture("43.png", 0, 0, 0, -1, 0, -29),
			new InputPicture("42.png", 0, 0, 0, -1, 0, -24),
			new InputPicture("41.png", 0, 0, 0, -1, 0, -19),
			new InputPicture("40.png", 0, 0, 0, -1, 0, -14),
			new InputPicture("39.png", 0, 0, 0, -1, 0, -9),
			new InputPicture("37.png", 0, 0, 0, -1, 0, -4),
			new InputPicture("36.png", 0, 0, 0, -1, 0, 0),
			new InputPicture("44.png", 0, 0, 0, -1, 0, 4),
			new InputPicture("45.png", 0, 0, 0, -1, 0, 9),
			new InputPicture("46.png", 0, 0, 0, -1, 0, 14),
			
		};
		
		/* compute the expected translation for the inputs
		 * rotate d and add c
		 * d = -1,-6,11
		 * 
		 * translate rvec to open-cv coords
		 * */
		
		/*the vector from the rotation point d to c.*/
		MatOfDouble dc = new MatOfDouble();
		dc.fromArray(new double[]{-1,-6,-11});
		
		MatOfDouble d = new MatOfDouble();
		d.fromArray(new double[]{1,-53,-11});
		
		/*d + dc = 0,-59,-22 = c*/
		
		for(InputPicture ip:inputs){
			/*non opencv rvec*/
			Mat rvec = new MatOfDouble();
			((MatOfDouble)rvec).fromArray(new double[]{ toRadians(ip.expectedRoll), toRadians(ip.expectedPitch), toRadians(ip.expectedYaw) });
			/*rvec is now a 3x3 matrix*/
			Calib3d.Rodrigues(rvec, rvec);
			Core.gemm(basisExchangeEuropeanCartesianToOpenCV, rvec, 1, new Mat(), 0, rvec);
			Core.gemm(rvec, basisExchangeOpenCVToEuropeanCartesian, 1, new Mat(), 0, rvec);
			/*rvec is now a 3x3 matrix in the open-cv coordinate system*/
			
			Mat rotatedDc = new Mat();
			/*rotate dc*/
			Core.gemm(rvec, dc, 1, new Mat(), 0, rotatedDc);
			
			Mat expectedTranslation = new Mat();
			/*expected translation in the open-cv space:*/
			Core.add(d, rotatedDc, expectedTranslation);
			
			/*translate the translation to cartesian*/
			Core.gemm(basisExchangeOpenCVToEuropeanCartesian, expectedTranslation, 1, new Mat(), 0, expectedTranslation);
			
			/*update the input-picture object*/
			for(int i=0;i<3;i++){
				double[] val = expectedTranslation.get(i, 0);
				switch(i){
					case 0:
						ip.dx = val[0];
						break;
					case 1:
						ip.dy = val[0];
						break;
					case 2:
						ip.dz = val[0];
						break;
				}
			}
			
			/*the input picture is now created successful - add it to the dataset by running the detector*/
			addToDataset(ip, dataset);
		}
		xmlInterface.storeDataset(dataset, xmlFile);
	}
	
	private static double toRadians(double degree) {
		return (degree * Math.PI) / 180.0;
	}

	public static void createDatasetTestenvironment1() throws FileNotFoundException, ParserConfigurationException, TransformerFactoryConfigurationError, TransformerException{
		Mat cameraMatrix = new Mat(new Size(3,3), CvType.CV_64F);
		for(int i=0;i<CAMERA_MATRIX_VALUES.length; i++){
			cameraMatrix.put(i, 0, CAMERA_MATRIX_VALUES[i]);
		};
		
		MatOfDouble distortionMatrix = new MatOfDouble();
		distortionMatrix.fromArray(DISTORTION_MATRIX_VALUES);
		
		File xmlFile = new File("ressources/input-data-test-environment-1-z.xml");
		Dataset dataset = new Dataset(new LinkedList<Picture>(), cameraMatrix, distortionMatrix);
		XmlInterface xmlInterface = new XmlInterface();
		
		Mat basisExchangeOpenCVToEuropeanCartesian = Mat.zeros(new Size(3,3),CvType.CV_64F);
		double[] row0 = new double[]{0,0,1};
		double[] row1 = new double[]{-1,0,0};
		double[] row2 = new double[]{0,1,0};
		basisExchangeOpenCVToEuropeanCartesian.put(0, 0, row0);
		basisExchangeOpenCVToEuropeanCartesian.put(1, 0, row1);
		basisExchangeOpenCVToEuropeanCartesian.put(2, 0, row2);
		
		Mat basisExchangeEuropeanCartesianToOpenCV = basisExchangeOpenCVToEuropeanCartesian.inv();
		
		/*
		 * open-cv coordinates
		 * c = 0,-59,-22 
		 * 
		 * */
		InputPicture[] inputs = new InputPicture[]{
				/*roll test series*/
				new InputPicture("61.png", 0, 0, 0, 0, 0, 0),
				new InputPicture("56.png", 0, 0, 0, 5, 0, 0),
				new InputPicture("57.png", 0, 0, 0, 10, 0, 0),
				new InputPicture("58.png", 0, 0, 0, 15, 0, 0),
				new InputPicture("59.png", 0, 0, 0, 20, 0, 0),
				new InputPicture("60.png", 0, 0, 0, 25, 0, 0),
				
				/*pitch test series*/
				new InputPicture("62.png", 0, 0, 0, 0, 0, 0),
				new InputPicture("63.png", 0, 0, 0, 0, 5, 0),
				new InputPicture("64.png", 0, 0, 0, 0, 10, 0),
				new InputPicture("65.png", 0, 0, 0, 0, 15, 0),
				new InputPicture("66.png", 0, 0, 0, 0, 20, 0),
				
				/*yaw test series*/
				new InputPicture("67.png", 0, 0, 0, 0, 0, -0),
				new InputPicture("68.png", 0, 0, 0, 0, 0, -5),
				new InputPicture("69.png", 0, 0, 0, 0, 0, -10),
				new InputPicture("70.png", 0, 0, 0, 0, 0, -15),
		};
		
		/* compute the expected translation for the inputs
		 * rotate d and add c
		 * d = -1,-6,11
		 * 
		 * translate rvec to open-cv coords
		 * */
		
		/*the vector from the rotation point d to c.*/
		MatOfDouble dc = new MatOfDouble();
		dc.fromArray(new double[]{-1,-6,-11});
		
		MatOfDouble d = new MatOfDouble();
		d.fromArray(new double[]{1,-53,-11});
		
		/*d + dc = 0,-59,-22 = c*/
		
		for(InputPicture ip:inputs){
			/*non opencv rvec*/
			MatOfDouble rvec = new MatOfDouble();
			rvec.fromArray(new double[]{ ip.expectedRoll, ip.expectedPitch, ip.expectedYaw });
			/*rvec is now a 3x3 matrix*/
			Calib3d.Rodrigues(rvec, rvec);
			Core.gemm(basisExchangeEuropeanCartesianToOpenCV, rvec, 1, new Mat(), 0, rvec);
			Core.gemm(basisExchangeOpenCVToEuropeanCartesian, rvec, 1, new Mat(), 0, rvec);
			/*rvec is now a 3x3 matrix in the open-cv coordinate system*/
			
			Mat rotatedDc = new Mat();
			/*rotate dc*/
			Core.gemm(rvec, dc, 1, new Mat(), 0, rotatedDc);
			
			Mat expectedTranslation = new Mat();
			/*expected translation in the open-cv space:*/
			Core.add(d, rotatedDc, expectedTranslation);
			
			/*translate the translation to cartesian*/
			Core.gemm(basisExchangeOpenCVToEuropeanCartesian, expectedTranslation, 1, new Mat(), 0, expectedTranslation);
			
			/*update the input-picture object*/
			for(int i=0;i<3;i++){
				double[] val = expectedTranslation.get(i, 0);
				switch(i){
				case 0:
					ip.dx = val[0];
					break;
				case 1:
					ip.dy = val[0];
					break;
				case 2:
					ip.dz = val[0];
					break;
				}
			}
			
			/*the input picture is now created successful - add it to the dataset by running the detector*/
			addToDataset(ip, dataset);
		}
		xmlInterface.storeDataset(dataset, xmlFile);
	}
	
	private static void addToDataset(InputPicture picture, Dataset dataset){
		String filename = picture.filename;
		double expectedDx = picture.dx;
		double expectedDy = picture.dy;
		double expectedDz = picture.dz;
		double expectedRoll = picture.expectedRoll;
		double expectedPitch = picture.expectedPitch;
		double expectedYaw = picture.expectedYaw;
		
		List<LED> leds = new LinkedList<LED>();
		/*detect leds*/
		Mat image = Highgui.imread("ressources/images/" + picture.filename);
		List<ColoredImagePatchMatch> detectedLEDs = headTracker.getDetectedLEDs(image);
		for(ColoredImagePatchMatch match:detectedLEDs){
			int led = match.getLed();
			Point imgCoord = match.getImagePatch().getCenter();
			/*get world coord*/
			Point3 worldCoordinate = trainingData.getWorldCoordinate(led);
			leds.add(new LED(led, imgCoord, worldCoordinate));
		}
		
		dataset.addPicture(new Picture(filename, leds, expectedDx, expectedDy, expectedDz, expectedRoll, expectedPitch, expectedYaw));
	}
	
	static class InputPicture{
		private String filename;
		private double dx;
		private double dy;
		private double dz;
		private double expectedRoll;
		private double expectedPitch;
		private double expectedYaw;
		public InputPicture(String filename, double dx, double dy, double dz,
				double expectedRoll, double expectedPitch, double expectedYaw) {
			this.filename = filename;
			this.dx = dx;
			this.dy = dy;
			this.dz = dz;
			this.expectedRoll = expectedRoll;
			this.expectedPitch = expectedPitch;
			this.expectedYaw = expectedYaw;
		}
	}
}
