import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFFont;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.HSSFColor;
import org.opencv.calib3d.Calib3d;
import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.MatOfPoint3f;
import org.opencv.core.Point;
import org.opencv.core.Point3;
import org.opencv.core.Scalar;
import org.xml.sax.SAXException;


public class TestSolvePnPMain {
	
	
	public static void main(String[] args) throws SAXException, IOException, ParserConfigurationException {
		testSolvePnp(new File("ressources/input-data.xml"), new File("ressources/output-data.xls"));
	}
	
	public static void testSolvePnp(File inputXmlFile, File outputXlsFile) throws SAXException, IOException, ParserConfigurationException{
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
		
		/*load dataset*/
		File xmlFile = inputXmlFile;
		XmlInterface xmlInterface = new XmlInterface();
		Dataset dataset = xmlInterface.loadDataset(xmlFile);
		
		/*open xls file*/
		HSSFWorkbook workbook = new HSSFWorkbook();
		HSSFCellStyle contentCellStyle = workbook.createCellStyle();
		contentCellStyle.setAlignment(HSSFCellStyle.ALIGN_LEFT);
		HSSFSheet worksheet = workbook.createSheet("solvePnP Test - " + inputXmlFile.getName() );
		/*create headers*/
		HSSFRow header = worksheet.createRow(0);
		
		HSSFFont headerFont = workbook.createFont();
		headerFont.setFontName(HSSFFont.FONT_ARIAL);
		headerFont.setFontHeightInPoints((short) 20);
		headerFont.setBoldweight(HSSFFont.BOLDWEIGHT_BOLD);
		headerFont.setColor(HSSFColor.BLACK.index);
		
		String[] headerFields = new String[]{
			"image-name",
			"detected-leds",
			
			"expected dx",
			"computed dx",
			"difference dx",
			"difference-median dx",
			"relative difference dx",
			
			
			"expected dy",
			"computed dy",
			"difference dy",
			"difference-median dy",
			"relative difference dy",
			
			"expected dz",
			"computed dz",
			"difference dz",
			"difference-median dz",
			"relative difference dz",
			
			"expected roll",
			"computed roll",
			"difference roll",
			"difference-median roll",
			"relative difference roll",
			
			"expected pitch",
			"computed pitch",
			"difference pitch",
			"difference-median pitch",
			"relative difference pitch",
			
			"expected yaw",
			"computed yaw",
			"difference yaw",
			"difference-median yaw",
			"relative difference yaw",
		};
		
		HSSFCellStyle style = workbook.createCellStyle();
		style.setFillPattern(HSSFCellStyle.BIG_SPOTS);
		style.setAlignment(HSSFCellStyle.ALIGN_CENTER);
		style.setWrapText(true);
		style.setBorderLeft(HSSFCellStyle.BORDER_HAIR);
		style.setBorderRight(HSSFCellStyle.BORDER_HAIR);
		style.setFont(headerFont);
		int collumnsCount = 0;
		for(String columnName:headerFields){
			worksheet.setColumnWidth(collumnsCount, 12000);
			HSSFCell cell = header.createCell(collumnsCount++);
			cell.setCellValue(columnName);
			cell.setCellType(HSSFCell.CELL_TYPE_STRING);
			cell.setCellStyle(style);
		}
		header.setHeight((short)450);
		
		/*compute row data*/
		int rowsCount = 1;
		int variableOffset = 5;
		double[] differences = new double[6];
		PosePicture[] posesToPicturs = new PosePicture[dataset.getPictures().size()];
		int i=0;
		int posesCount = 0;
		for(Picture picture:dataset.getPictures()){
			PosePicture poseToPicture = new PosePicture(picture, null);
			posesToPicturs[i++] = poseToPicture;
			System.out.println("compute solvePnP for:" + picture.getFilename());
			
			HSSFRow row = worksheet.createRow(rowsCount++);
			HSSFCell c = row.createCell(0);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getFilename());
			
			StringBuilder sbLeds = new StringBuilder();
			for(LED led:picture.getLeds()){
				sbLeds.append(led.getNumber());
				sbLeds.append(", ");
			}
			if(sbLeds.length() > 0){
				sbLeds.replace(sbLeds.length()-2, sbLeds.length(), "");
				sbLeds.reverse();
			}
			c = row.createCell(1);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(sbLeds.toString());
			
			

			int entry = 2;
			c = row.createCell(entry);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getExpectedDx());
			
			c = row.createCell(entry+variableOffset*1);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getExpectedDy());
			
			c = row.createCell(entry+variableOffset*2);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getExpectedDz());
			
			
			c = row.createCell(entry+variableOffset*3);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getExpectedRoll());
			
			c = row.createCell(entry+variableOffset*4);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getExpectedPitch());
			
			
			c = row.createCell(entry+variableOffset*5);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(picture.getExpectedYaw());
			
			
			Mat rvec = new MatOfDouble();
			Mat tvec = new MatOfDouble();
			
			((MatOfDouble)rvec).fromArray(new double[]{0,0,0});
			((MatOfDouble)tvec).fromArray(new double[]{0,59, 22});
			List<LED> leds = picture.getLeds();
			Point3[] objectPointsArr	= new Point3[leds.size()];
			Point[] imagePointsArr		= new Point[leds.size()];
			int j=0;
			for(LED led:leds){
				objectPointsArr[j] = led.getWorldCoordinate();
				imagePointsArr[j] = led.getImageCoordinate();
				j++;
			}
			
			MatOfPoint3f objectPoints = new MatOfPoint3f();
			MatOfPoint2f imagePoints = new MatOfPoint2f();
			
			objectPoints.fromArray(objectPointsArr);
			imagePoints.fromArray(imagePointsArr);
			
			System.out.println(picture.getFilename());
			System.out.println(objectPoints.dump());
			System.out.println(imagePoints.dump());
			System.out.println(dataset.getCameraMatrix().dump());
			System.out.println(dataset.getDistortionMatrix().dump());
			if(objectPoints.height() < 4){
				continue;
			}else if(objectPoints.height() == 4){
//				Calib3d.solvePnP(objectPoints, imagePoints, dataset.getCameraMatrix(), dataset.getDistortionMatrix(), rvec, tvec);
				Calib3d.solvePnP(
					objectPoints, imagePoints,
					dataset.getCameraMatrix(), dataset.getDistortionMatrix(),
					rvec, tvec,
					true,
					Calib3d.CV_ITERATIVE
				);
			}else{
//				Calib3d.solvePnPRansac(objectPoints, imagePoints, dataset.getCameraMatrix(), dataset.getDistortionMatrix(), rvec, tvec);
				Calib3d.solvePnPRansac(
				/*
				 * MatOfPoint3f objectPoints, MatOfPoint2f imagePoints,
				 * Mat cameraMatrix, MatOfDouble distCoeffs,
				 * Mat rvec, Mat tvec,
				 * boolean useExtrinsicGuess,
				 * int iterationsCount,
				 * float reprojectionError,
				 * int minInliersCount,
				 * Mat inliers,
				 * int flags
				 * */
				objectPoints, imagePoints,
				dataset.getCameraMatrix(), dataset.getDistortionMatrix(),
				rvec, tvec,
				true,
				10, //iterations count
				3, //reprojection error,
				5,
				new Mat(),
				Calib3d.ITERATIVE
			);
			}
			/*generate outputs*/
			
			/*compute real translation and rotation data in the world coordinate system*/
			Mat rotation = new Mat();
			Mat negRotation = new Mat();
			Mat translation = new Mat();
			
			Calib3d.Rodrigues(rvec, rotation);
			
			rotation = rotation.t();
			Core.multiply(rotation, new Scalar(-1), negRotation);
			Core.gemm(negRotation, tvec, 1, new Mat(), 0, translation);
			
			Pose pose = new Pose(rotation, translation);
			poseToPicture.pose = pose;
			posesCount++;
			
			entry = 3;
			
			c = row.createCell(entry);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(pose.getDx());
			
			c = row.createCell(entry+variableOffset*1);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(pose.getDy());
			
			c = row.createCell(entry+variableOffset*2);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(pose.getDz());
			
			
			c = row.createCell(entry+variableOffset*3);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(pose.getRoll());
			
			c = row.createCell(entry+variableOffset*4);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(pose.getPitch());
			
			
			c = row.createCell(entry+variableOffset*5);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(pose.getYaw());
			
			
			/*difference*/
			entry = 4;
			double difference = picture.getExpectedDx() - pose.getDx();
			differences[0] += difference;
			c = row.createCell(entry);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference);
			
			difference = picture.getExpectedDy() - pose.getDy();
			differences[1] += difference;
			c = row.createCell(entry+variableOffset*1);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference);
			
			difference = picture.getExpectedDz() - pose.getDz();
			differences[2] += difference;
			c = row.createCell(entry+variableOffset*2);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference);
			
			
			difference = picture.getExpectedRoll() - pose.getRoll();
			differences[3] += difference;
			c = row.createCell(entry+variableOffset*3);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference);
			
			difference = picture.getExpectedPitch() - pose.getPitch();
			differences[4] += difference;
			c = row.createCell(entry+variableOffset*4);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference);
			
			
			difference = picture.getExpectedYaw() - pose.getYaw();
			differences[5] += difference;
			c = row.createCell(entry+variableOffset*5);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference);
			
			/*image name*/
			/*visible led's*/
			/*6x expected value | computed value | difference | median_difference | expected_with_median_difference | computed_value | difference */
		}
		for(int j=0;j<differences.length;j++){
			differences[j] /= posesCount;
		}
		
		rowsCount=1;
		for(PosePicture poseToPicture:posesToPicturs){
			HSSFRow row = worksheet.getRow(rowsCount++);
			HSSFCell c;
			Picture picture = poseToPicture.picture;
			Pose pose = poseToPicture.pose;
			if(pose==null){
				continue;
			}
			
			/*median*/
			int entry = 5;
			c = row.createCell(entry);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(differences[0]);
			
			c = row.createCell(entry+variableOffset*1);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(differences[1]);
			
			c = row.createCell(entry+variableOffset*2);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(differences[2]);
			
			
			c = row.createCell(entry+variableOffset*3);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(differences[3]);
			
			c = row.createCell(entry+variableOffset*4);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(differences[4]);
			
			
			c = row.createCell(entry+variableOffset*5);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(differences[5]);
			
			/*relative difference*/
			entry = 6;
			double difference = picture.getExpectedDx() - pose.getDx();
			c = row.createCell(entry);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference-differences[0]);
			
			difference = picture.getExpectedDy() - pose.getDy();
			c = row.createCell(entry+variableOffset*1);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference-differences[1]);
			
			difference = picture.getExpectedDz() - pose.getDz();
			c = row.createCell(entry+variableOffset*2);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference-differences[2]);
			
			
			difference = picture.getExpectedRoll() - pose.getRoll();
			c = row.createCell(entry+variableOffset*3);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference-differences[3]);
			
			difference = picture.getExpectedPitch() - pose.getPitch();
			c = row.createCell(entry+variableOffset*4);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference-differences[4]);
			
			
			difference = picture.getExpectedYaw() - pose.getYaw();
			c = row.createCell(entry+variableOffset*5);
			c.setCellStyle(contentCellStyle);
			c.setCellValue(difference-differences[5]);
			
		}
		
		
		/*output xls file*/
		FileOutputStream fileOut = new FileOutputStream(outputXlsFile);
		workbook.write(fileOut);
		fileOut.flush();
		fileOut.close();
	}
	
	static class PosePicture{
		Picture picture;
		Pose pose;
		public PosePicture(Picture picture, Pose pose) {
			super();
			this.picture = picture;
			this.pose = pose;
		}
	}
}
