package head_tracker;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.opencv.calib3d.Calib3d;
import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.MatOfFloat;
import org.opencv.core.MatOfInt;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.MatOfPoint3f;
import org.opencv.core.Point;
import org.opencv.core.Point3;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import de.tu_darmstadt.fsr.headtracker.TrainingData;

//TODO compare to... http://www.morethantechnical.com/2010/11/10/20-lines-ar-in-opencv-wcode/

public abstract class HeadTracker {
//	protected final double[] DISTORTION_MATRIX_VALUES = new double[]{
//			/*This matrix should have 5 values*/
//			0.04569467373955304,
//			0.1402980385369059,
//			0,
//			0,
//			-0.2982135315849994
//	};
	
//	protected final double[] DISTORTION_MATRIX_VALUES = new double[]{
//			/*This matrix should have 5 values*/
//			0.08245931646421553,
//			-0.9893762277047577,
//			0,
//			0,
//			3.23553287438898
//	};

	protected final double[] DISTORTION_MATRIX_VALUES = new double[]{
			/*This matrix should have 5 values*/
			0.07444480392067945,
			-0.7817175834131075,
			0,
			0,
			2.65433773093283
	};
//	protected final double[] DISTORTION_MATRIX_VALUES = new double[]{
//			/*This matrix should have 5 values*/
//			0.08909941096327206,
//			-0.9537960457721699,
//			0,
//			0,
//			3.449728790843752
//	};
//	
//	protected final double[] DISTORTION_MATRIX_VALUES = new double[]{
//			/*This matrix should have 5 values*/
//			0.0,
//			0.0,
//			0,
//			0,
//			0
//	};
	protected final double[][] CAMERA_MATRIX_VALUES = new double[][]{
			/*This matrix should have 3x3 values*/
//			{748.6595405553738, 0, 319.5},
//			{0, 748.6595405553738, 239.5},
//			{0, 0, 1}
//			{698.1744297982436, 0, 320},
//			{0, 698.1744297982436, 240},
//			{0, 0, 1}
//			{707.1226937511951, 0, 319.5},
//			{0, 707.1226937511951, 239.5},
//			{0, 0, 1}
//			{702.1458656346429, 0, 319.5},
//			{0, 702.1458656346429, 239.5},
//			{0, 0, 1}
			{702.1458656346429, 0, 319.5},
			{0, 702.1458656346429, 239.5},
			{0, 0, 1}
	};
	//http://helpdesk.metaio.com/questions/30643/how-to-do-hand-eye-calibration-moverio-bt-200
	
	/**
	 * trained data for LED recognition
	 */
	protected TrainingData trainedData;
	
	protected MatOfDouble distortionMatrix;
	protected Mat cameraMatrix;

	public HeadTracker(Object... params){
		init(params);
		initDestortionMatrix();
		initCameraMatrix();
		trainedData = loadTrainedLEDData();
	}
	
	protected abstract void init(Object... params);
	
	private void initDestortionMatrix(){
		distortionMatrix = new MatOfDouble();
		distortionMatrix.fromArray(DISTORTION_MATRIX_VALUES);
	}
	
	private void initCameraMatrix(){
		cameraMatrix = new Mat(new Size(3,3), CvType.CV_64F);
		for(int i=0;i<CAMERA_MATRIX_VALUES.length; i++){
			cameraMatrix.put(i, 0, CAMERA_MATRIX_VALUES[i]);
		}
	}
	
	/**
	 * Loads trained data for each led.
	 * This data includes:
	 *  - the trained color histograms
	 *  - 2D bounding boxes for possible LED neighbors
	 */
	protected abstract TrainingData loadTrainedLEDData();
	
	public Pose getCameraPose(Mat imageBGR){
		MatOfPoint3f ledWorldCoordinates = new MatOfPoint3f();
		MatOfPoint2f ledImageCoordinates  = new MatOfPoint2f();
		detectLEDs(imageBGR, ledWorldCoordinates, ledImageCoordinates);
		
		System.out.println("world-coordinates: " + ledWorldCoordinates.dump());
		System.out.println("led-coordinates: " + ledImageCoordinates.dump());
		
		if(ledWorldCoordinates.height() < 4){
			return null;
		}
		
		/**
		 * output vector rotation and translation
		 * could used as input if intrinsic guess is set to true
		 * */
		Mat rVec = new MatOfDouble();
		Mat tVec = new MatOfDouble();
		
		((MatOfDouble)rVec).fromArray(new double[]{0,0,0});
		((MatOfDouble)tVec).fromArray(new double[]{0,59, 22});
		/*
		 * The algorithmens have been implemented by Alexander Shishkov 
		 * http://answers.opencv.org/question/258/is-anyone-using-p3p-or-epnp-in-solvepnp/
		 * */
		if(true && (ledWorldCoordinates.height() == 4 || ledImageCoordinates.height() == 4)){
			Calib3d.solvePnP(
					ledWorldCoordinates, ledImageCoordinates,
					cameraMatrix, distortionMatrix,
					rVec, tVec,
					true,
					Calib3d.CV_ITERATIVE
			);
		}else{
			System.out.println("ransac");
			Mat inliners = new Mat();
//			ledImageCoordinates = MatOfPoint2f.fromNativeAddr(ledImageCoordinates.rowRange(0, 4).nativeObj);
//			ledWorldCoordinates = MatOfPoint3f.fromNativeAddr(ledWorldCoordinates.rowRange(0, 4).nativeObj);
//			System.out.println(ledImageCoordinates.dump());
//			System.out.println(ledWorldCoordinates.dump());
			//TODO include intrinsic guess by gyroscope
			//TODO adjust ransac parameter
			Calib3d.solvePnPRansac(
					/*
					 * MatOfPoint3f objectPoints, MatOfPoint2f imagePoints,
					 * Mat cameraMatrix, MatOfDouble distCoeffs,
					 * Mat rvec, Mat tvec,
					 * boolean useExtrinsicGuess,
					 * int iterationsCount,
					 * float reprojectionError,
					 * int minInliersCount,
					 * Mat inliers,
					 * int flags
					 * */
					ledWorldCoordinates, ledImageCoordinates,
					cameraMatrix, distortionMatrix,
					rVec, tVec,
					true,
					10, //iterations count
					5, //reprojection error,
					5,
					inliners,
					Calib3d.CV_ITERATIVE
					);
					System.out.println("inliners-count:" + inliners.size());
		}
		
		
		/*
		 * http://stackoverflow.com/questions/18637494/camera-position-in-world-coordinate-from-cvsolvepnp#_=_
		 * 
		 * (http://stackoverflow.com/questions/14444433/calculate-camera-world-position-with-opencv-python)
		 * (http://stackoverflow.com/questions/14515200/python-opencv-solvepnp-yields-wrong-translation-vector?lq=1)
		 * */
		
		/*
		 * homography approach:
		 * http://gotoanswer.com/?q=OpenCV+%2B+OpenGL%3A+proper+camera+pose+using+solvePnP
		 */

		System.out.println("original-rvec-dump:" + rVec.dump());
		System.out.println("original-tvec-dump:" + tVec.dump());
		Mat rotation = new Mat();
		Mat negRotation = new Mat();
		Mat translation = new Mat();
		
		Calib3d.Rodrigues(rVec, rotation);
		System.out.println("rotation: " + rotation.dump());
		
		rotation = rotation.t();
		Core.multiply(rotation, new Scalar(-1), negRotation);
		Core.gemm(negRotation, tVec, 1, new Mat(), 0, translation);
		System.out.println("translation: " + translation.dump());
		
		
		Pose pose = new Pose(rotation, translation);
		/*
		 * validate pose
		 * 
		 * Assume, that in the environment of the FSR the translation and rotation will never be zero.
		 * */
		Mat zeros = Mat.zeros(new Size(1,3), CvType.CV_64F);
		pose.setValid(!zeros.equals(rVec) || zeros.equals(tVec));
		
		return pose;
	}
	
	protected List<ColoredImagePatchMatch> computeLEDDetection(Mat imageBGR){
		/*compute a value intensity based mask to detect possible LEDs*/
		Mat hsv = new Mat(imageBGR.size(), imageBGR.type());
		Imgproc.cvtColor(imageBGR, hsv, Imgproc.COLOR_BGR2HSV);
		List<Mat> hsvSplit = new ArrayList<Mat>(3);
		Core.split(hsv, hsvSplit);
		Mat val = hsvSplit.get(2);
		Mat valMask = computeValMask(val);
		
		/* 
		 * compute  image patches, containing the LED silhouette in the RGB space
		 * ... and ...
		 * compute a color histograms for each patch and compare it with the trained data
		 * */
		Mat rgb = new Mat();
		Imgproc.cvtColor(imageBGR, rgb, Imgproc.COLOR_BGR2RGB);

		List<ColoredImagePatchMatch>[] matches = computeSilhouetteColoredImagePatcheMatches(valMask, rgb);
		
		/*build a graph to find the longest detectable LED chain with the highest detection possibility*/
		List<ColoredImagePatchMatch> detectedLEDs = computeGraphBasedLEDDetection(matches);
		
		return detectedLEDs;
	}
	
	/**
	 * @param ledImageCoordinates 
	 * @param ledWorldCoordinates 
	 * @return A Mat[] containing the 3D Points of the detected LEDs and the corrosponding image Coordinates
	 * 	Mat[0] contains the 3D-Coordinates
	 * 	Mat[1] contains the corrososponding image coordinates
	 * */
	private void detectLEDs(Mat imageBGR, MatOfPoint3f ledWorldCoordinates, MatOfPoint2f ledImageCoordinates){
		
		List<ColoredImagePatchMatch> detectedLEDs = computeLEDDetection(imageBGR);
		
		/*collect and store 3D world and 2D image coordinates of the detected LEDs*/
		Point3[] worldCoordinates = new Point3[detectedLEDs.size()];
		Point[] imageCoordinates = new Point[detectedLEDs.size()];
		int i=0;
		for(ColoredImagePatchMatch led:detectedLEDs){
			imageCoordinates[i] = led.getImagePatch().getCenter();
			worldCoordinates[i] = trainedData.getWorldCoordinate(led.getLed());
			i++;
		}
		
		System.out.println("----printing world coordinates ---- ");
		for(int j=0; j<=8;j++){
			Point3 worldCoordinate = trainedData.getWorldCoordinate(j);
			System.out.println("led: " + j + " -> " + worldCoordinate.x + "," + worldCoordinate.y + "," + worldCoordinate.z);
		}
		System.out.println(ledWorldCoordinates);
		System.out.println(worldCoordinates.length);
		ledWorldCoordinates.fromArray(worldCoordinates);
		ledImageCoordinates.fromArray(imageCoordinates);
	}
	
	private List<ColoredImagePatchMatch> computeGraphBasedLEDDetection(List<ColoredImagePatchMatch>[] matches){
		Node q = new Node(null);
		Node s = new Node(null);
		int nodesCount = 2;
		
		@SuppressWarnings("unchecked")
		List<Node>[] nodes = new LinkedList[matches.length];
		for(int i=0; i < matches.length; i++){
			nodes[i] = new LinkedList<Node>();
			for(ColoredImagePatchMatch match:matches[i]){
				nodes[i].add(new Node(match));
				nodesCount++;
			}
		}
		
		List<Edge> edges = new LinkedList<Edge>();
		
		/*iterate threw all matches and build the graph*/
		for(List<Node> list:nodes){
			for(Node nodeMatch:list){
				ColoredImagePatchMatch match = nodeMatch.getMatch();
				
				edges.add(new Edge( q, nodeMatch, match.getDistance() ));
				edges.add(new Edge( nodeMatch, s, 0 ));
				
				int neighborR = match.getLed() +1;
				
				
				if(neighborR < trainedData.getLEDsCount()){
					for(Node nodeNeighbor:nodes[neighborR]){
						ColoredImagePatchMatch neighbor = nodeNeighbor.getMatch();
						if(isValidNeigbor(match, neighbor)){
							edges.add(new Edge( nodeMatch, nodeNeighbor, neighbor.getDistance() ));
						}
					}
				}
			}
		}
		return findShortestPath(edges, nodesCount, q, s);
	}
	
	private List<ColoredImagePatchMatch> findShortestPath(List<Edge> edges, int nodesCount, Node a, Node b) {
		a.setDistance(0);
		
		for(int i=0; i<nodesCount-1; i++){
			for(Edge edge:edges){
				Node source = edge.getSource();
				Node destination = edge.getDestination();
				double newDistance = source.getDistance() + edge.getCosts();
				if( newDistance < destination.getDistance() ){
					destination.setDistance(newDistance);
					destination.setPrevious(source);
				}
			}
		}
		
		for(Edge edge:edges){
			Node source = edge.getSource();
			Node destination = edge.getDestination();
			double newDistance = source.getDistance() + edge.getCosts();
			if( newDistance < destination.getDistance() ){
				throw new RuntimeException("negative cycle found");
			}
		}
		
		List<ColoredImagePatchMatch> path = new LinkedList<ColoredImagePatchMatch>();
		Node current = b;
		while(current != a && current!=null){
			if(current != b){
				path.add(current.getMatch());
			}
			current = current.getPrevious();
		}
		return path;
	}

	private boolean isValidNeigbor(ColoredImagePatchMatch match, ColoredImagePatchMatch neighbor){
		/*An LED can not be its own neighbor*/
		if(match.getLed() == neighbor.getLed()){
			return false;
		}
		/*Allow only one steps for a neighborhood*/
		if(Math.abs(match.getLed()-neighbor.getLed())>1){
			return false;
		}
		
		/*Only allow right neighbors*/
		boolean movingRight = match.getLed()-neighbor.getLed() < 0;
		if(!movingRight){
			return false;
		}
		
		/*Only allow neighbors in the right distance area*/
		int minXDist = trainedData.getMinDistX()[match.getLed()];
		int minYDist = trainedData.getMinDistY()[match.getLed()];
		int maxXDist = trainedData.getMaxDistX()[match.getLed()];
		int maxYDist = trainedData.getMaxDistY()[match.getLed()];
		double procentualBoundingBoxStretch = 100;
		
		/*perform bounding-box stretch*/
		int boundingBoxDistX = Math.abs(maxXDist - minXDist);
		int boundingBoxDistY = Math.abs(maxYDist - minYDist);
		int stretchX = (int) ((boundingBoxDistX/200.0) * procentualBoundingBoxStretch);
		int stretchY = (int) ((boundingBoxDistY/200.0) * procentualBoundingBoxStretch);
		minXDist -= stretchX;
		minYDist -= stretchY;
		maxXDist += stretchX;
		maxYDist += stretchY;
		
		org.opencv.core.Point centerA = match.getImagePatch().getCenter();
		org.opencv.core.Point centerB = neighbor.getImagePatch().getCenter();
		int distX = (int) (centerB.x - centerA.x);
		int distY = (int) (centerB.y - centerA.y);
		
		if( 
			distX < minXDist ||
			distX > maxXDist ||
			distY < minYDist ||
			distY > maxYDist 
		){
			return false;
		}
		
		return true;
	}
	
	private  List<ColoredImagePatchMatch>[] computeSilhouetteColoredImagePatcheMatches(Mat valMask, Mat rgb){
		
		List<MatOfPoint> contours = new LinkedList<MatOfPoint>();
		Imgproc.findContours(valMask, contours, new Mat(), Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);
		
		Mat silhouettes = computeSilhouettes(valMask);

		/*smooth rgb image*/
		Imgproc.GaussianBlur(rgb, rgb, new Size(3,3), 0.8);
		
		computeRGBIntensityNormalization(rgb);
		
		/*wrap input images for the histogram computation in a list*/
		List<Mat> rgbSplit = new ArrayList<Mat>(3);
		Core.split(rgb, rgbSplit);
		
		List<Mat> rgbR = new ArrayList<Mat>(1);
		rgbR.add(rgbSplit.get(0));
		List<Mat> rgbG = new ArrayList<Mat>(1);
		rgbG.add(rgbSplit.get(1));
		List<Mat> rgbB = new ArrayList<Mat>(1);
		rgbB.add(rgbSplit.get(2));
		
		@SuppressWarnings("unchecked")
		List<ColoredImagePatchMatch>[] matches = new LinkedList[trainedData.getLEDsCount()];
		for(int i=0;i<matches.length; i++){
			matches[i] = new LinkedList<ColoredImagePatchMatch>();
		}
		
		for(MatOfPoint contour:contours){
			Mat histR = new Mat();
			Mat histG = new Mat();
			Mat histB = new Mat();
			
			MatOfPoint2f  contourMoP2f = new MatOfPoint2f( contour.toArray() );
			org.opencv.core.Point center = new org.opencv.core.Point();
			Imgproc.minEnclosingCircle(contourMoP2f, center, null);
			
			/*compute a silhouette mask, containing the silhouette area in a bounding box around the center of the contour*/
			Mat contourSilhouetteMask = Mat.zeros(valMask.size(), valMask.type());
			
			double x = center.x;
			double y = center.y;
			
			double xTL = Math.max(0, x-25);
			double yTL = Math.max(0, y-25);
			double xBR = Math.min(contourSilhouetteMask.width()-1,x+25);					;
			double yBR = Math.min(contourSilhouetteMask.height()-1,y+25);

			Core.rectangle(contourSilhouetteMask, new org.opencv.core.Point(xTL, yTL), new org.opencv.core.Point(xBR, yBR), new Scalar(255,255,255), -1);
			Core.bitwise_and(contourSilhouetteMask, silhouettes, contourSilhouetteMask);

			/*compute histograms*/
			MatOfInt channels = new MatOfInt(0);
			MatOfInt histSize = new MatOfInt(32);
			MatOfFloat ranges = new MatOfFloat(0, 1f);
			
			Imgproc.calcHist(rgbR, channels, contourSilhouetteMask, histR, histSize, ranges, false);
			Imgproc.calcHist(rgbG, channels, contourSilhouetteMask, histG, histSize, ranges, false);
			Imgproc.calcHist(rgbB, channels, contourSilhouetteMask, histB, histSize, ranges, false);
			
			Core.normalize(histR, histR, 0, 1, Core.NORM_MINMAX);
			Core.normalize(histG, histG, 0, 1, Core.NORM_MINMAX);
			Core.normalize(histB, histB, 0, 1, Core.NORM_MINMAX);
			
			ColoredImagePatch imagePatch = new ColoredImagePatch(center, histR, histG, histB);
			/*compare histogram with trained histograms*/
			
			Mat[][] hists = trainedData.getHists();
			
			for(int led=0; led<hists.length; led++){
				double distance = computeDistanceValue(imagePatch, led);
				if(distance < 1.3 ){
					//assume a positive match
					matches[led].add(new ColoredImagePatchMatch(led, imagePatch, distance));
				}
			}
		}
		
		
		return matches;
	}
	
	private double computeDistanceValue(ColoredImagePatch imagePatch, int led){
		double result =   Imgproc.compareHist(imagePatch.getHistR(), trainedData.getHists()[led][0], Imgproc.CV_COMP_BHATTACHARYYA)
						+ Imgproc.compareHist(imagePatch.getHistG(), trainedData.getHists()[led][1], Imgproc.CV_COMP_BHATTACHARYYA)
						+ Imgproc.compareHist(imagePatch.getHistB(), trainedData.getHists()[led][2], Imgproc.CV_COMP_BHATTACHARYYA);
		return result;
	}
	
	
	private void computeRGBIntensityNormalization(Mat rgb){
		List<Mat> rgbSplit = new ArrayList<Mat>(3);
		Core.split(rgb, rgbSplit);
		Mat intensity = Mat.zeros(rgb.size(), CvType.CV_16U);
		Core.add(intensity, rgbSplit.get(0), intensity, new Mat(), CvType.CV_16U);
		Core.add(intensity, rgbSplit.get(1), intensity, new Mat(), CvType.CV_16U);
		Core.add(intensity, rgbSplit.get(2), intensity, new Mat(), CvType.CV_16U);
		Core.divide(rgbSplit.get(0), intensity, rgbSplit.get(0), 1, CvType.CV_32F);
		Core.divide(rgbSplit.get(1), intensity, rgbSplit.get(1), 1, CvType.CV_32F);
		Core.divide(rgbSplit.get(2), intensity, rgbSplit.get(2), 1, CvType.CV_32F);
		Core.merge(rgbSplit, rgb);
	}
	
	private Mat computeValMask(Mat val){
		Mat mask = new Mat();
		Core.bitwise_and(computeValCockpitMask(val), computeValLEDMask(val), mask);
		Imgproc.erode(mask, mask, Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(2,2)));
		return mask;
	}
	
	private Mat computeSilhouettes(Mat valMask){
		Mat silhouettes = new Mat();
		Imgproc.dilate(valMask, silhouettes, Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(17,17)));
		Core.bitwise_not(valMask, silhouettes, silhouettes);
		return silhouettes;
	}
	
	private Mat computeValCockpitMask(Mat val){
		Mat mask = new Mat();
		Core.inRange(val, new Scalar(0),  new Scalar(70), mask);
		Imgproc.erode(mask, mask, Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(9,5)));
		
		/*This kernel draws lines at the bottom bigger*/
		Mat kernel = Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(40,10));
		kernel.rowRange(5, 10).setTo(new Scalar(0));
		
		Imgproc.dilate(mask, mask, kernel);
		
		/*
		 * find large contour with the lowest points
		 */
		List<MatOfPoint> contours = new LinkedList<MatOfPoint>();
		Imgproc.findContours(mask, contours, new Mat(), Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);
		MatOfPoint lowestBigContour = null;
		double lowestPoint = 0;
		for(MatOfPoint contour:contours){
			for(org.opencv.core.Point p:contour.toList()){
				if(p.y>lowestPoint){
					if(Imgproc.contourArea(contour)>4000){
						lowestPoint = p.y;
						lowestBigContour = contour;
					}
				}
			}
		}
		contours.clear();
		contours.add(lowestBigContour);
		if(lowestBigContour!=null){
			mask = Mat.zeros(mask.size(), mask.type());
			Imgproc.drawContours(mask, contours, 0, new Scalar(255,255,255),  -1);
		}else{
			mask = new Mat(mask.size(), mask.type(), new Scalar(255));
		}

		/*blur the mask a little*/
		Imgproc.dilate(mask, mask, Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(100,1)));
		return mask;
	}
	
	private Mat computeValLEDMask(Mat val){
		Mat mask = new Mat();
		
		Core.inRange(val, new Scalar(245),  new Scalar(255), mask);
		
		Imgproc.erode(mask, mask, Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(3,2)));
		Imgproc.dilate(mask, mask, Imgproc.getStructuringElement(Imgproc.MORPH_RECT, new Size(8,8)));
		
		
		List<MatOfPoint> contours = new LinkedList<MatOfPoint>();
		Imgproc.findContours(mask, contours, new Mat(), Imgproc.RETR_EXTERNAL, Imgproc.CHAIN_APPROX_SIMPLE);
		Iterator<MatOfPoint> contourIterator = contours.iterator();
		while(contourIterator.hasNext()){
			MatOfPoint contour = contourIterator.next();
			if(Imgproc.contourArea(contour) > 1300){
				contourIterator.remove();
			}
		}
		
		mask = Mat.zeros(mask.size(), mask.type());
		Imgproc.drawContours(mask, contours, -1, new Scalar(255), -1);
		
		return mask;
	}
}